@@ -4,9 +4,6 @@
  * only implements in-memory history.
  */
 
-#if !defined(lint) && !defined(no_RCSids)
-static char *RCSid = "$Id: history.c,v 1.5 1994/06/17 19:52:41 michael Exp michael $";
-#endif
 /*
  *	This file contains
  *	a)	the original in-memory history  mechanism
@@ -19,19 +16,24 @@ static char *RCSid = "$Id: history.c,v 1.5 1994/06/17 19:52:41 michael Exp micha
  */
 
 #include "sh.h"
+#include "ksh_stat.h"
 
-#ifdef EASY_HISTORY
+#ifdef HISTORY
+# ifdef EASY_HISTORY
 
-# ifndef HISTFILE
-#  define HISTFILE ".pdksh_hist"
-# endif
+#  ifndef HISTFILE
+#   ifdef OS2
+#    define HISTFILE "history.ksh"
+#   else /* OS2 */
+#    define HISTFILE ".pdksh_hist"
+#   endif /* OS2 */
+#  endif
 
-#else
+# else
 /*	Defines and includes for the complicated case */
 
-# include "ksh_stat.h"
-# include <sys/file.h>
-# include <sys/mman.h>
+#  include <sys/file.h>
+#  include <sys/mman.h>
 
 /*
  *	variables for handling the data file
@@ -47,14 +49,16 @@ static void histinsert ARGS((Source *, int, unsigned char *));
 static void writehistfile ARGS((int, char *));
 static int sprinkle ARGS((int));
 
-# ifdef MAP_FILE
-#  define MAP_FLAGS	(MAP_FILE|MAP_PRIVATE)
-# else
-#  define MAP_FLAGS	MAP_PRIVATE
-# endif
+#  ifdef MAP_FILE
+#   define MAP_FLAGS	(MAP_FILE|MAP_PRIVATE)
+#  else
+#   define MAP_FLAGS	MAP_PRIVATE
+#  endif
 
-#endif	/* of EASY_HISTORY */
+# endif	/* of EASY_HISTORY */
 
+static char   **histget ARGS((char *str, int approx));
+static void	histbackup ARGS((void));
 static char    *histrpl ARGS((char *s, char *pat, char *rep, int global));
 
 static char   **current;	/* current postition in history[] */
@@ -69,48 +73,40 @@ c_fc(wp)
 {
 	register char *id;
 	struct shf *shf;
-	FILE *f;
 	struct temp UNINITIALIZED(*tf);
 	register char **hp;
 	char **hbeg, **hend;
 	char *p, *cmd = NULL;
 	int lflag = 0, nflag = 0, sflag = 0, rflag = 0, gflag = 0;
-	int done = 0;
-	void histbackup();
+	int optc;
 
-	for (wp++; (id = *wp) != NULL && *id++ == '-' && !done; wp++)
-		while (*id && !done) {
-			switch (*id++) {
-			  case 'l':
-				lflag++;
-				break;
-			  case 'n':
-				nflag++;
-				break;
-			  case 'r':
-				rflag++;
-				break;
-			  case 'g':
-				gflag++;
-				break;
-			  case 'e':
-				if (++wp && (p = *wp)) {
-					if (p[0] == '-' && !p[1]) {
-						sflag++;
-					} else {
-						cmd = strnsave(p, strlen(p) + 4, ATEMP);
-						strcat(cmd, " $_");
-					}
-				} else
-					errorf("argument expected\n");
-				id = "";
-				break;
-			  default:
-				wp--;
-				done++;
-				break;
+	while ((optc = ksh_getopt(wp, &builtin_opt, "e:glnr")) != EOF)
+		switch (optc) {
+		  case 'e':
+			p = builtin_opt.optarg;
+			if (strcmp(p, "-") == 0)
+				sflag++;
+			else {
+				cmd = strnsave(p, strlen(p) + 4, ATEMP);
+				strcat(cmd, " $_");
 			}
+			break;
+		  case 'g': /* non-at&t ksh */
+			gflag++;
+			break;
+		  case 'l':
+			lflag++;
+			break;
+		  case 'n':
+			nflag++;
+			break;
+		  case 'r':
+			rflag++;
+			break;
+		  case '?':
+			return 1;
 		}
+	wp += builtin_opt.optind;
 
 	if (sflag) {
 		char *pat = NULL, *rep = NULL;
@@ -123,30 +119,31 @@ c_fc(wp)
 				rep = p;
 				*rep++ = '\0';
 			} else
-				hp = histget(id);
+				hp = histget(id, FALSE);
 		}
 
-		if (hp == NULL || hp < history)
-			errorf("cannot find history\n");
-		if (pat == NULL)
-			strcpy(line, *hp);
-		else
-			histrpl(*hp, pat, rep, gflag);
+		if (hp == NULL || hp < history) {
+			/* XXX improve error */
+			bi_errorf("cannot find history");
+			return 1;
+		}
+		if (pat == NULL) {
+			int len = strlen(*hp);
+			p = strnsave(*hp, len + 1, source->areap);
+			p[len] = '\n';
+			p[len + 1] = '\0';
+		} else if (!(p = histrpl(*hp, pat, rep, gflag)))
+			return 1;
 		histbackup();
-#ifdef EASY_HISTORY
-		histsave(line); 
-#else
-		histsave(source->line+1, line, 1);
-#endif
-		histpush--; 
-		line[0] = '\0';
+		histsave(source->line+1, p, 1);
+		histpush--;
 		return 0;
 	}
 
 	if (*wp != NULL) {
-		hbeg = histget(*wp++); /* first */
+		hbeg = histget(*wp++, TRUE); /* first */
 		if (*wp != NULL)
-			hend = histget(*wp++); /* last */
+			hend = histget(*wp++, TRUE); /* last */
 		else if (lflag)
 			hend = (histptr - hbeg > 16) ? hbeg + 16 : histptr;
 		else
@@ -159,8 +156,11 @@ c_fc(wp)
 		if (hbeg < history)
 			hbeg = history;
 	}
-	if (hbeg == NULL || hend == NULL)
-		errorf("can't find history\n");
+	if (hbeg == NULL || hend == NULL) {
+		/* XXX improve error */
+		bi_errorf("can't find history");
+		return 1;
+	}
 
 	if (lflag)
 		shf = shf_fdopen(1, SHF_WR, (struct shf *) 0);
@@ -169,45 +169,93 @@ c_fc(wp)
 		tf = maketemp(ATEMP);
 		tf->next = e->temps; e->temps = tf;
 		shf = shf_open(tf->name, O_WRONLY|O_CREAT|O_TRUNC, 0666, 0);
-		if (shf == NULL)
-			errorf("cannot create temp file %s\n", tf->name);
+		if (shf == NULL) {
+			bi_errorf("cannot create temp file %s", tf->name);
+			return 1;
+		}
 	}
 
 	for (hp = (rflag ? hend : hbeg); rflag ? (hp >= hbeg) : (hp <= hend);
 	      rflag ? hp-- : hp++) {
-		if (!nflag)
-			shf_fprintf(shf, "%3d: ",
-				source->line - (int)(histptr-hp));
-		shf_fprintf(shf, "%s\n", *hp);
+		if (lflag) {
+			char *s, *t;
+
+			if (!nflag)
+				shf_fprintf(shf, "%d",
+					source->line - (int)(histptr-hp));
+			shf_putchar('\t', shf);
+			/* print multi-line commands correctly */
+			for (s = *hp; (t = strchr(s, '\n')); s = t)
+				shf_fprintf(shf, "%.*s\t", ++t - s, s, null);
+			shf_fprintf(shf, "%s\n", s);
+		} else
+			shf_fprintf(shf, "%s\n", *hp);
 	}
 
 	shf_flush(shf);
 	if (lflag)
 		return 0;
-	else
-		shf_close(shf);
+	shf_close(shf);
 
 	setstr(local("_"), tf->name);
-	if (cmd) {
-		command(cmd); /* edit temp file */
-		afree(cmd, ATEMP);
-	} else
-		command("${FCEDIT:-/bin/ed} $_");
 
-	f = fopen(tf->name, "r");
-	if (f == NULL)
-		errorf("cannot open temp file %s\n", tf->name);
-	/* we push the editted lines onto the history list */
-	while (fgets(line, sizeof(line), f) != NULL) {
-#ifdef EASY_HISTORY
-		histsave(line); 
-#else
-		histsave(source->line, line, 1); 
-#endif
-		histpush--; 
+	/* XXX: source should not get trashed by this.. */
+	{
+		Source *sold = source;
+
+		if (cmd)
+			command(cmd); /* edit temp file */
+		else
+			command("${FCEDIT:-/bin/ed} $_");
+		source = sold;
+	}
+
+	{
+		struct stat statb;
+		XString xs;
+		char *xp, *p;
+		int n;
+		int first = 1;
+
+		if (!(shf = shf_open(tf->name, O_RDONLY, 0666, 0))) {
+			bi_errorf("cannot open temp file %s", tf->name);
+			return 1;
+		}
+
+		histbackup();
+		n = fstat(shf_fileno(shf), &statb) < 0 ? 128
+			: statb.st_size + 1;
+		Xinit(xs, xp, n, source->areap);
+		while (1) {
+			/* read a line */
+			xp = Xstring(xs, xp);
+			while ((p = shf_getse(xp, Xnleft(xs, xp), shf))) {
+				xp = p;
+				if (Xstring(xs, xp) != xp && p[-1] == '\n')
+					break;
+				XcheckN(xs, xp, Xlength(xs, xp));
+			}
+			if (shf_error(shf)) {
+				bi_errorf("error reading temp file %s - %s",
+					tf->name, strerror(shf_errno(shf)));
+				shf_close(shf);
+				return 1;
+			}
+			if (!p && xp == Xstring(xs, xp))
+				break;
+			strip_nuls(Xstring(xs, xp), Xlength(xs, xp));
+# ifdef EASY_HISTORY
+			if (!first) {
+				histappend(Xstring(xs, xp), 1); 
+				continue;
+			}
+			first = 0;
+# endif
+			histpush--;
+			histsave(source->line+1, Xstring(xs, xp), 1);
+		}
+		shf_close(shf);
 	}
-	line[0] = '\0';
-	fclose(f);
 
 	return 0;
 }
@@ -215,7 +263,7 @@ c_fc(wp)
 /******************************/
 /* Back up over last histsave */
 /******************************/
-void
+static void
 histbackup()
 {
 	static int last_line = -1;
@@ -232,26 +280,26 @@ histbackup()
  * get pointer to history given pattern
  * pattern is a number or string
  */
-char **
-histget(str)
+static char **
+histget(str, approx)
 	char *str;
+	int approx;
 {
 	register char **hp = NULL;
+	int n;
 
-	if (*str == '-')
-		hp = histptr + getn(str);
-	else
-	if (digit(*str))
-		hp = histptr + (getn(str) - source->line);
-	else 
-	if (*str == '?') {	/* unanchored match */
-		for (hp = histptr-1; hp >= history; hp--)
-			if (strstr(*hp, str+1) != NULL)
-				break;
-	} else {		/* anchored match */
-		for (hp = histptr; hp >= history; hp--)
-			if (strncmp(*hp, str, strlen(str)) == 0)
-				break;
+	if (getn(str, &n)) {
+		if (n < 0) {
+			hp = histptr + n;
+			if (approx && hp < history)
+				hp = history;
+		} else
+			hp = histptr + (n - source->line);
+	} else {
+		int anchored = *str == '?' ? (++str, 0) : 1;
+
+		hp = &history[findhist((histptr - 1) - history, 0, str,
+					anchored)];
 	}
 
 	return (history <= hp && hp <= histptr) ? hp : NULL;
@@ -263,31 +311,37 @@ histrpl(s, pat, rep, global)
 	char *pat, *rep;
 	int global;
 {
-	char *s1, *p, *last = NULL;
-	int len = strlen(pat);
+	char *s1, *last = NULL;
+	int pat_len = strlen(pat);
 	int rep_len = strlen(rep);
+	XString xs;
+	char *xp;
 
-	line[0] = '\0';
-	p = line;
+	Xinit(xs, xp, 128, source->areap);
 	while ((s1 = strstr(s, pat))) {
-		if ((p - line) + (s1 - s) + rep_len >= LINE)
-			errorf("substitution too long\n");
-		strncpy(p, s, s1 - s);		/* first part */
-		p += s1 - s;
-		strcpy(p, rep);	/* replacement */
-		p += rep_len;
-		s = s1 + len;
-		last = s1;
+		XcheckN(xs, xp, (s1 - s) + rep_len);
+		memcpy(xp, s, s1 - s);		/* first part */
+		xp += s1 - s;
+		memcpy(xp, rep, rep_len);	/* replacement */
+		xp += rep_len;
+		s = last = s1 + pat_len;
 		if (!global)
-			s = "";
+			break;
 	}
 	if (last) {
-		if (p - line + strlen(last + len) >= LINE)
-			errorf("substitution too long\n");
-		strcpy(p, last + len);		/* last part */
-	} else
-		errorf("substitution failed\n");
-	return line;
+		int last_len = strlen(last);
+
+		XcheckN(xs, xp, last_len);
+		memcpy(xp, last, last_len);
+		xp += last_len;
+		*xp++ = '\n';
+		*xp++ = '\0';
+	} else {
+		bi_errorf("substitution failed");
+		Xfree(xs, xp);
+		return (char *) 0;
+	}
+	return Xclose(xs, xp);
 }
 
 /*
@@ -328,26 +382,25 @@ histnum(n)
  * direction.
  */
 int
-findhist(start, fwd, str)
+findhist(start, fwd, str, anchored)
 	int	start;
 	int	fwd;
 	char 	*str;
+	int	anchored;
 {
 	char	**hp;
 	int	maxhist = histptr - history;
+	int	incr = fwd ? 1 : -1;
+	int	len = strlen(str);
 
 	if (start < 0 || start >= maxhist)
 		start = maxhist;
 
-	if (fwd) {
-		for (hp = &history[start]; *hp != (char *) 0; hp++)
-			if (strstr(*hp, str))
-				return hp - history;
-	} else {
-		for (hp = &history[start]; hp >= history; hp--)
-			if (strstr(*hp, str))
-				return hp - history;
-	}
+	hp = &history[start];
+	for (; hp >= history && hp <= histptr; hp += incr)
+		if ((anchored && strncmp(*hp, str, len) == 0)
+		    || (!anchored && strstr(*hp, str)))
+			return hp - history;
 
 	return -1;
 }
@@ -360,17 +413,19 @@ void
 sethistsize(n)
 	int n;
 {
-	int	offset;
-	
-	if (n != histsize) {
-		offset = histptr - history;
+	if (n > 0 && n != histsize) {
+		int cursize = histptr - history;
+
+		/* save most recent history */
+		if (n < cursize) {
+			memmove(history, histptr - n, n * sizeof(char *));
+			cursize = n;
+		}
+
 		history = (char **)aresize(history, n*sizeof(char *), APERM);
 
-		if (n < histsize && offset > histsize)
-			offset = histsize;
-
 		histsize = n;
-		histptr = history + offset;
+		histptr = history + cursize;
 	}
 }
 
@@ -394,12 +449,12 @@ sethistfile(name)
 	/*
 	 * its a new name - possibly
 	 */
-#ifdef EASY_HISTORY
+# ifdef EASY_HISTORY
 	if (hname) {
 		afree(hname, APERM);
 		hname = NULL;
 	}
-#else
+# else
 	if (histfd) {
 		/* yes the file is open */
 		(void) close(histfd);
@@ -411,7 +466,7 @@ sethistfile(name)
 		histptr = history - 1;
 		source->line = 0;
 	}
-#endif
+# endif
 
 	hist_init(source);
 }
@@ -423,27 +478,30 @@ void
 init_histvec()
 {
 	if (history == (char **)NULL) {
-		histsize = HISTORY;
+		histsize = HISTORYSIZE;
 		history = (char **)alloc(histsize*sizeof (char *), APERM);
-		histptr = history-1;
+		histptr = history - 1;
 	}
 }
 
-#ifdef EASY_HISTORY
+# ifdef EASY_HISTORY
 /*
  * save command in history
  */
 void
-histsave(cmd)
+histsave(lno, cmd, dowrite)
+	int lno;	/* ignored (compatibility with COMPLEX_HISTORY) */
 	char *cmd;
+	int dowrite;	/* ignored (compatibility with COMPLEX_HISTORY) */
 {
 	register char **hp = histptr;
 	char *cp;
 
 	if (++hp >= history + histsize) { /* remove oldest command */
-		afree((void*)*history, APERM);
-		for (hp = history; hp < history + histsize - 1; hp++)
-			hp[0] = hp[1];
+		afree((void*)history[0], APERM);
+		memmove(history, history + 1,
+			sizeof(history[0]) * (histsize - 1));
+		hp = &history[histsize - 1];
 	}
 	*hp = strsave(cmd, APERM);
 	/* trash trailing newline but allow imbedded newlines */
@@ -498,19 +556,19 @@ hist_init(s)
 	hstarted = 1;
 
 	if ((f = strval(global("HISTFILE"))) == NULL || *f == '\0') {
-#if 1 /* Don't use history file unless the user asks for it */
+# if 1 /* Don't use history file unless the user asks for it */
 		hname = NULL;
 		return;
-#else
+# else
 		char *home = strval(global("HOME"));
 		int len;
 
 		if (home == NULL)
-			home = "";
+			home = null;
 		f = HISTFILE;
 		hname = alloc(len = strlen(home) + strlen(f) + 2, APERM);
 		shf_snprintf(hname, len, "%s/%s", home, f);
-#endif
+# endif
 	} else
 		hname = strsave(f, APERM);
 
@@ -518,24 +576,26 @@ hist_init(s)
 		int pos = 0, nread = 0;
 		int contin = 0;		/* continuation of previous command */
 		char *end;
+		char hline[LINE + 1];
 
 		while (1) {
 			if (pos >= nread) {
 				pos = 0;
-				nread = fread(line, 1, LINE, fh);
+				nread = fread(hline, 1, LINE, fh);
 				if (nread <= 0)
 					break;
-				line[nread] = '\0';
+				hline[nread] = '\0';
 			}
-			end = strchr(line + pos, 0); /* will always succeed */
+			end = strchr(hline + pos, 0); /* will always succeed */
 			if (contin)
-				histappend(line + pos, 0);
-			else
-				histsave(line + pos);
-			pos = end - line + 1;
-			contin = end == &line[nread];
+				histappend(hline + pos, 0);
+			else {
+				s->line++;
+				histsave(0, hline + pos, 0);
+			}
+			pos = end - hline + 1;
+			contin = end == &hline[nread];
 		}
-		line[0] = '\0';
 		fclose(fh);
 	}
 }
@@ -565,13 +625,13 @@ hist_finish()
     hp = history;
   if (hname && (fh = fopen(hname, "w")))
   {
-    for (i = 0; i < histsize && hp[i]; i++)
+    for (i = 0; hp + i <= histptr && hp[i]; i++)
       fprintf(fh, "%s%c", hp[i], '\0');
     fclose(fh);
   }
 }
 
-#else /* EASY_HISTORY */
+# else /* EASY_HISTORY */
 
 /*
  *	Routines added by Peter Collinson BSDI(Europe)/Hillside Systems to
@@ -632,9 +692,9 @@ histsave(lno, cmd, dowrite)
  *	Each command is
  *	<command byte><command number(4 bytes)><bytes><null>
  */
-#define HMAGIC1		0xab
-#define HMAGIC2		0xcd
-#define COMMAND		0xff
+# define HMAGIC1		0xab
+# define HMAGIC2		0xcd
+# define COMMAND		0xff
 
 void
 hist_init(s)
@@ -642,7 +702,6 @@ hist_init(s)
 {
 	unsigned char	*base;
 	int	lines;
-	int	bytes;
 	int	fd;
 	
 	if (Flag(FTALKING) == 0)
@@ -768,7 +827,7 @@ hist_shrink(oldbase, oldbytes)
 	/*
 	 *	create temp file
 	 */
-	(void) shf_snprintf(nfile, sizeof(nfile), "%s.%d", hname, getpid());
+	(void) shf_snprintf(nfile, sizeof(nfile), "%s.%d", hname, procpid);
 	if ((fd = creat(nfile, 0600)) < 0)
 		return 1;
 
@@ -811,21 +870,20 @@ hist_skip_back(base, bytes, no)
 	register int lines = 0;
 	register unsigned char *ep;
 
-	
-
-	for (ep = base + *bytes; ep > base; ep--)
-	{
-		while (*ep != COMMAND) {
-			if (--ep == base)
-				break;
-		}
+	for (ep = base + *bytes; --ep > base; ) {
+		/* this doesn't really work: the 4 byte line number that is
+		 * encoded after the COMMAND byte can itself contain the
+		 * COMMAND byte....
+		 */
+		for (; ep > base && *ep != COMMAND; ep--)
+			;
+		if (ep == base)
+			break;
 		if (++lines == no) {
 			*bytes = *bytes - ((char *)ep - (char *)base);
 			return ep;
 		}
 	}
-	if (ep > base)
-		return base;
 	return NULL;
 }
 
@@ -986,4 +1044,29 @@ sprinkle(fd)
 	return(write(fd, mag, 2) != 2);
 }
 
-#endif
+# endif
+#else /* HISTORY */
+
+/* No history to be compiled in: dummy routines to avoid lots more ifdefs */
+void
+init_histvec()
+{
+}
+void
+hist_init(s)
+	Source *s;
+{
+}
+void
+hist_finish()
+{
+}
+void
+histsave(lno, cmd, dowrite)
+	int lno;
+	char *cmd;
+	int dowrite;
+{
+	errorf("history not enabled");
+}
+#endif /* HISTORY */
