@@ -3,8 +3,10 @@
 
 #if defined(DIRENT) || defined(_POSIX_VERSION)
 # include <dirent.h>
+# define NLENGTH(dirent)	(strlen(dirent->d_name))
 #else
 # define dirent direct
+# define NLENGTH(dirent)	(dirent->d_namlen)
 # ifdef SYSNDIR
 #  include <sys/ndir.h>
 # endif /* SYSNDIR */
