@@ -47,3 +47,16 @@ typedef int WAIT_T;
 #  define WSTOPSIG(s)		(((s) >> 8) & 0xff)
 # endif
 #endif /* HAVE_UNION_WAIT */
+
+#if !defined(HAVE_WAITPID) && defined(HAVE_WAIT3)
+#  define ksh_waitpid(p, s, o)	wait3((s), (o), (struct rusage *) 0)
+#else /* !HAVE_WAITPID && HAVE_WAIT3 */
+# ifdef HAVE_UNION_WAIT
+   /* The cast here is because most POSIX BSD systems have union wait,
+    * but pass an int * to waitpid() since POSIX says it takes an int *.
+    */
+#  define ksh_waitpid(p, s, o)	waitpid((p), (int *) (s), (o))
+# else /* HAVE_UNION_WAIT */
+#  define ksh_waitpid(p, s, o)	waitpid((p), (s), (o))
+# endif /* HAVE_UNION_WAIT */
+#endif /* !HAVE_WAITPID && HAVE_WAIT3 */
