@@ -2,15 +2,10 @@
  * Miscellaneous functions
  */
 
-#if !defined(lint) && !defined(no_RCSids)
-static char *RCSid = "$Id: misc.c,v 1.2 1994/05/19 18:32:40 michael Exp michael $";
-#endif
-
 #include "sh.h"
 #ifdef HAVE_LIMITS_H
 # include <limits.h>
 #endif
-#include "expand.h"
 
 #ifndef UCHAR_MAX
 # define UCHAR_MAX	0xFF
@@ -53,7 +48,8 @@ initctypes()
 	setctypes(" \t\n|&;<>()", C_LEX1); /* \0 added automatically */
 	setctypes("*@#!$-?", C_VAR1);
 	setctypes(" \t\n", C_IFSWS);
-	setctypes("=-+?#%", C_SUBOP);
+	setctypes("=-+?", C_SUBOP1);
+	setctypes("#%", C_SUBOP2);
 }
 
 /* convert unsigned long to base N string */
@@ -102,51 +98,67 @@ strnsave(s, n, ap)
 	return strncat(ns, s, n);
 }
 
+/* called from expand.h:XcheckN() to grow buffer */
+char *
+Xcheck_grow_(xsp, xp, more)
+	XString *xsp;
+	char *xp;
+	int more;
+{
+	char *old_beg = xsp->beg;
+
+	xsp->len += more > xsp->len ? more : xsp->len;
+	xsp->beg = aresize(xsp->beg, xsp->len + 8, xsp->areap);
+	xsp->end = xsp->beg + xsp->len;
+	return xsp->beg + (xp - old_beg);
+}
+
 struct option options[] = {
-	/* Special cases (see parse_args()): -A, -c, -o, -s */
-	/* options are sorted by their longnames */
+	/* Special cases (see parse_args()): -A, -o, -s.
+	 * Options are sorted by their longnames - the order of these
+	 * entries MUST match the order of sh_flag F* enumerations in sh.h.
+	 */
 	{ "allexport",	'a',		OF_ANY },
 #ifdef BRACEEXPAND
 	{ "braceexpand",  0,		OF_ANY }, /* non-standard */
 #endif
-	{ "bgnice",	  0,		OF_ANY }, /* not done */
-	{ "",	 	'c',		     0 }, /* -c on command line */
-#ifdef VI
-	{ "vi-tabcomplete",  0, 	OF_ANY },
-#endif
+	{ "bgnice",	  0,		OF_ANY },
+	{ null,	 	'c',	    OF_CMDLINE },
 #ifdef EMACS
 	{ "emacs",	  0,		OF_ANY },
 #endif
 	{ "errexit",	'e',		OF_ANY },
 #ifdef EMACS
-	{ "gmacs",	  0,		OF_ANY }, /* not done */
+	{ "gmacs",	  0,		OF_ANY },
 #endif
 	{ "ignoreeof",	  0,		OF_ANY },
 	{ "interactive",'i',	    OF_CMDLINE },
 	{ "keyword",	'k',		OF_ANY },
-	{ "markdirs",	'X',		OF_ANY }, /* not done */
+	{ "markdirs",	'X',		OF_ANY },
 #ifdef JOBS
 	{ "monitor",	'm',		OF_ANY },
 #else /* JOBS */
-	{ "",		'm',		     0 }, /* so FMONITOR not ifdef'd */
+	{ null,		'm',		     0 }, /* so FMONITOR not ifdef'd */
 #endif /* JOBS */
 	{ "noclobber",	'C',		OF_ANY },
 	{ "noexec",	'n',		OF_ANY },
 	{ "noglob",	'f',		OF_ANY },
-	{ "nolog",	  0,		OF_ANY }, /* not done */
+	{ "nolog",	  0,		OF_ANY }, /* no effect */
 #ifdef	JOBS
 	{ "notify",	'b',		OF_ANY },
 #endif	/* JOBS */
 	{ "nounset",	'u',		OF_ANY },
-	{ "posix",	  0,		OF_ANY },
-	{ "privileged",	'p',		OF_ANY }, /* not done */
-	{ "restricted",	'r',	    OF_CMDLINE }, /* not done */
+	{ "posix",	  0,		OF_ANY }, /* non-standard */
+	{ "privileged",	'p',		OF_ANY },
+	{ "restricted",	'r',	    OF_CMDLINE },
 	{ "stdin",	's',	    OF_CMDLINE }, /* pseudo non-standard */
 	{ "trackall",	'h',		OF_ANY },
 	{ "verbose",	'v',		OF_ANY },
 #ifdef VI
 	{ "vi",		  0,		OF_ANY },
-	{ "viraw",	  0,		OF_ANY }, /* ignored */
+	{ "viraw",	  0,		OF_ANY }, /* no effect */
+	{ "vi-show8",	  0,		OF_ANY }, /* non-standard */
+	{ "vi-tabcomplete",  0, 	OF_ANY }, /* non-standard */
 #endif
 	{ "xtrace",	'x',		OF_ANY },
 	{ NULL,		  0,		     0 }
@@ -168,6 +180,30 @@ option(n)
 	return -1;
 }
 
+struct options_info {
+	int opt_width;
+	struct {
+		char	*name;
+		int	flag;
+	} opts[NELEM(options)];
+};
+
+/* format a single select menu item */
+static char *
+options_fmt_entry(arg, i, buf, buflen)
+	void *arg;
+	int i;
+	char *buf;
+	int buflen;
+{
+	struct options_info *oi = (struct options_info *) arg;
+
+	shf_snprintf(buf, buflen, "%-*s %s",
+		oi->opt_width, oi->opts[i].name,
+		Flag(oi->opts[i].flag) ? "on" : "off");
+	return buf;
+}
+
 static void
 printoptions(verbose)
 	int verbose;
@@ -175,18 +211,28 @@ printoptions(verbose)
 	int i;
 
 	if (verbose) {
-		/* verbose (at&t) version */
+		struct options_info oi;
+		int n, len;
+
+		/* verbose version */
 		shprintf("Current option settings\n");
-		for (i = 0; options[i].name; i++)
-			if (options[i].name[0])
-				shprintf("%-16s %s\n", options[i].name,
-					Flag(i) ? "on" : "off");
+
+		for (i = n = oi.opt_width = 0; options[i].name; i++)
+			if (options[i].name[0]) {
+				len = strlen(options[i].name);
+				oi.opts[n].name = options[i].name;
+				oi.opts[n++].flag = i;
+				if (len > oi.opt_width)
+					oi.opt_width = len;
+			}
+		print_columns(shl_stdout, n, options_fmt_entry, &oi,
+			      oi.opt_width + 5);
 	} else {
 		/* short version */
 		for (i = 0; options[i].name; i++)
 			if (Flag(i) && options[i].name[0])
 				shprintf("%s ", options[i].name);
-		shprintf("\n");
+		shprintf(newline);
 	}
 }
 
@@ -240,44 +286,53 @@ change_flag(f, what, newval)
 # endif /* EMACS */
 			Flag(f) = newval;
 		}
-	}
+	} else
 #endif /* EDIT */
 	/* Turning off -p? */
 	if (f == FPRIVILEGED && oldval && !newval) {
+#ifdef OS2
+		;
+#else /* OS2 */
 		setuid(getuid());
 		setgid(getgid());
+#endif /* OS2 */
+	} else if (f == FPOSIX && newval) {
+#ifdef BRACEEXPAND
+		Flag(FBRACEEXPAND) = 0
+#endif /* BRACEEXPAND */
+		;
 	}
 }
 
 /* parse command line & set command arguments.  returns the index of
- * non-option arguments
+ * non-option arguments, -1 if there is an error.
  */
 int
-parse_args(argv, what, cargp, setargsp)
+parse_args(argv, what, setargsp)
 	char **argv;
 	int	what;		/* OF_CMDLINE or OF_SET */
-	char	**cargp;
 	int	*setargsp;
 {
-	static char cmd_opts[NELEM(options) + 5]; /* c:o:\0 */
-	static char set_opts[NELEM(options) + 6]; /* A:o;s\0 */
+	static char cmd_opts[NELEM(options) + 3]; /* o:\0 */
+	static char set_opts[NELEM(options) + 5]; /* Ao;s\0 */
 	char *opts;
+	char *arrayname;
 	Getopt go;
-	int i, optc, set, sortargs = 0;
+	int i, optc, set, sortargs = 0, arrayset = 0;
 
 	/* First call?  Build option strings... */
 	if (cmd_opts[0] == '\0') {
 		char *p;
 
 		/* c is also in options[], but it needs a trailing : */
-		strcpy(cmd_opts, "c:o:"); /* see cmd_opts[] declaration */
+		strcpy(cmd_opts, "o:"); /* see cmd_opts[] declaration */
 		p = cmd_opts + strlen(cmd_opts);
 		for (i = 0; options[i].name; i++)
 			if (options[i].c && (options[i].flags & OF_CMDLINE))
 				*p++ = options[i].c;
 		*p = '\0';
 
-		strcpy(set_opts, "A:o;s"); /* see set_opts[] declaration */
+		strcpy(set_opts, "Ao;s"); /* see set_opts[] declaration */
 		p = set_opts + strlen(set_opts);
 		for (i = 0; options[i].name; i++)
 			if (options[i].c && (options[i].flags & OF_CMDLINE))
@@ -290,21 +345,11 @@ parse_args(argv, what, cargp, setargsp)
 	while ((optc = ksh_getopt(argv, &go, opts)) != EOF) {
 		set = (go.info & GI_PLUS) ? 0 : 1;
 		switch (optc) {
-		case 'A':
-			if (!*go.optarg || *skip_varname(go.optarg, FALSE))
-				errorf("%s: is not an identifier\n", go.optarg);
-			set_array(go.optarg, set, argv + go.optind);
-			for (; argv[go.optind]; go.optind++)
-				;
-			go.info |= GI_DONE;	/* kludge to get out of loop */
+		  case 'A':
+			arrayset = set ? 1 : -1;
 			break;
 
-		case 'c':
-			Flag(FCOMMAND) = 1;
-			*cargp = go.optarg;
-			break;
-
-		case 'o':
+		  case 'o':
 			if (go.optarg == (char *) 0) {
 				/* lone -o: print options
 				 *
@@ -316,22 +361,18 @@ parse_args(argv, what, cargp, setargsp)
 				break;
 			}
 			i = option(go.optarg);
-			/* temporary hack to ease transition to braceexpand */
-			if (strcmp(go.optarg, "alternations") == 0) {
-				i = (int) FBRACEEXPAND;
-				/*
-				shellf("set: warning: use `%s' instead of `%s'\n",
-					options[i].name, go.optarg);
-				*/
-			}
 			if (i >= 0 && (options[i].flags & what))
 				change_flag((enum sh_flag) i, what, set);
-			else
-				errorf("%s: %s: bad option\n",
-					argv[0], go.optarg);
+			else {
+				bi_errorf("%s: bad option", go.optarg);
+				return -1;
+			}
 			break;
 
-		default:
+		  case '?':
+			return -1;
+
+		  default:
 			/* -s: sort positional params (at&t ksh stupidity) */
 			if (what == OF_SET && optc == 's') {
 				sortargs = 1;
@@ -345,9 +386,10 @@ parse_args(argv, what, cargp, setargsp)
 						    set);
 					break;
 				}
-			if (!options[i].name)
-				errorf("%s: parse_args: internal error (%c)\n",
-					argv[0], optc);
+			if (!options[i].name) {
+				internal_errorf(1, "parse_args: `%c'", optc);
+				return -1; /* not reached */
+			}
 		}
 	}
 	if (!(go.info & GI_MINUSMINUS) && argv[go.optind]
@@ -362,56 +404,69 @@ parse_args(argv, what, cargp, setargsp)
 	}
 	if (setargsp)
 		/* -- means set $#/$* even if there are no arguments */
-		*setargsp = (go.info & GI_MINUSMINUS) || argv[go.optind];
+		*setargsp = !arrayset && ((go.info & GI_MINUSMINUS)
+					  || argv[go.optind]);
 
+	if (arrayset) {
+		arrayname = argv[go.optind++];
+		if (!arrayname) {
+			bi_errorf("-A: missing array name\n");
+			return -1;
+		}
+		if (!*arrayname || *skip_varname(arrayname, FALSE)) {
+			bi_errorf("%s: is not an identifier\n",
+				arrayname);
+			return -1;
+		}
+	} else
+		arrayname = (char *) 0;	/* keep gcc happy */
 	if (sortargs) {
 		for (i = go.optind; argv[i]; i++)
 			;
 		qsortp((void **) &argv[go.optind], (size_t) (i - go.optind),
 			xstrcmp);
 	}
+	if (arrayset) {
+		set_array(arrayname, arrayset, argv + go.optind);
+		for (; argv[go.optind]; go.optind++)
+			;
+	}
 
 	return go.optind;
 }
 
-/* atoi with error detection */
+/* parse a decimal number: returns 0 if string isn't a number, 1 otherwise */
 int
-getn(as)
+getn(as, ai)
 	char *as;
+	int *ai;
 {
 	register char *s;
 	register int n;
+	int sawdigit = 0;
 
 	s = as;
 	if (*s == '-')
 		s++;
-	for (n = 0; digit(*s); s++)
-		n = (n*10) + (*s-'0');
-	if (*s)
-		errorf("%s: bad number\n", as);
-	return (*as == '-') ? -n : n;
+	for (n = 0; digit(*s); s++, sawdigit = 1)
+		n = n * 10 + (*s - '0');
+	*ai = (*as == '-') ? -n : n;
+	if (*s || !sawdigit)
+		return 0;
+	return 1;
 }
 
-/* like getn_ but uses shellf instead of errorf */
+/* getn() that prints error */
 int
-getn_(as, who)
+bi_getn(as, ai)
 	char *as;
-	char *who;
+	int *ai;
 {
-	register char *s;
-	register int n;
+	int rv = getn(as, ai);
 
-	s = as;
-	if (*s == '-')
-		s++;
-	for (n = 0; digit(*s); s++)
-		n = (n*10) + (*s-'0');
-	if (*s) {
-		shellf("%s: %s: bad number\n", who, as);
-		shf_flush(shl_out);
-		n = 1; /* ensure non-zero */
-	}
-	return (*as == '-') ? -n : n;
+	if (!rv)
+		bi_errorf("%s: bad number", as);
+	return rv;
 }
 
 /* -------- gmatch.c -------- */
@@ -433,8 +488,8 @@ gmatch(s, p)
 	if (s == NULL || p == NULL)
 		return 0;
 	while ((pc = *p++) != 0) {
-		sc = *s++;
-		if (pc ==  MAGIC)
+		sc = FILECHCONV(*s++);
+		if (pc == MAGIC)
 			switch (*p++) {
 			  case '[':
 				if (sc == 0 || (p = cclass(p, sc)) == NULL)
@@ -454,6 +509,16 @@ gmatch(s, p)
 				} while (*s++ != '\0');
 				return (0);
 
+#if 0
+			  /* [!+*?@](pattern|pattern|..) */
+			  case '!': /* matches none of the patterns */
+			  case '+': /* matches one or more times */
+			  case '*': /* matches zero or more times */
+			  case '?': /* matches zero or once */
+			  case '@': /* matches once */
+				/* fall through */
+#endif /* 0 */
+
 			  default:
 				if (sc != p[-1])
 					return 0;
@@ -472,6 +537,7 @@ cclass(p, sub)
 	register int sub;
 {
 	register int c, d, not, found = 0;
+	char *orig_p = p;
 
 	if ((not = (*p == MAGIC && *++p == NOT)))
 		p++;
@@ -479,18 +545,26 @@ cclass(p, sub)
 		if (*p == MAGIC)
 			p++;
 		if (*p == '\0')
-			return NULL;
-		c = *p;
-		if (p[1] == '-' && p[2] != ']') {
-			d = p[2];
-			p++;
+			/* No closing ] - act as if the opening [ was quoted */
+			return sub == '[' ? orig_p : NULL;
+		c = *p++;
+		if (p[0] == MAGIC && p[1] == '-'
+		    && (p[2] != MAGIC || p[3] != ']'))
+		{
+			p += 2; /* MAGIC- */
+			if (*p == MAGIC)
+				p++;
+			d = *p++;
+			/* POSIX says this is an invalid expression */
+			if (c > d)
+				return NULL;
 		} else
 			d = c;
 		if (c == sub || (c <= sub && sub <= d))
 			found = 1;
-	} while (*++p != ']');
+	} while (!(p[0] == MAGIC && p[1] == ']'));
 
-	return (found != not) ? p+1 : NULL;
+	return (found != not) ? p+2 : NULL;
 }
 
 /* -------- qsort.c -------- */
@@ -608,12 +682,13 @@ ksh_getopt_reset(go, flags)
 	go->buf[1] = '\0';
 }
 
+
 /* getopt() used for shell built-in commands, the getopts command, and
  * command line options.
  * A leading ':' in options means don't print errors, instead return '?'
  * or ':' and set go->optarg to the offending option character.
  * If GF_ERROR is set (and option doesn't start with :), errors result in
- * a call to errorf().
+ * a call to bi_errorf().
  *
  * Non-standard features:
  *	- ';' is like ':' in options, except the argument is optional
@@ -630,7 +705,8 @@ ksh_getopt_reset(go, flags)
  *	  Used for 'typeset -LZ4'.
  *	- accepts +c as well as -c IF the GF_PLUSOPT flag is present.  If an
  *	  option starting with + is accepted, the GI_PLUS flag will be set
- *	  in go->info.
+ *	  in go->info.  Once a - or + has been seen, all other options must
+ *	  start with the same character.
  */
 int
 ksh_getopt(argv, go, options)
@@ -647,39 +723,39 @@ ksh_getopt(argv, go, options)
 	if (go->info & GI_DONE)
 		return EOF;
 	if (go->p == 0 || (c = argv[go->optind - 1][go->p]) == '\0') {
+		char *arg = argv[go->optind], flag = arg ? *arg : '\0';
+
 		go->p = 1;
-		go->info &= ~GI_PLUS;
-		if (argv[go->optind] == (char *) 0
-		    || (argv[go->optind][0] != '-'
-			&& (!(go->flags & GF_PLUSOPT)
-			   || argv[go->optind][0] != '+'))
-		    || (c = argv[go->optind][1]) == '\0')
+		if (flag == '-' && arg[1] == '-' && arg[2] == '\0') {
+			go->optind++;
+			go->info |= GI_DONE | GI_MINUSMINUS;
+			return EOF;
+		}
+		if (arg == (char *) 0
+		    || ((flag != '-' || (go->info & GI_PLUS))
+			&& (!(go->flags & GF_PLUSOPT) || (go->info & GI_MINUS)
+			    || flag != '+'))
+		    || (c = arg[1]) == '\0')
 		{
 			go->info |= GI_DONE;
 			return EOF;
 		}
 		go->optind++;
-		if (c == '-' && !argv[go->optind - 1][2]) {
-			go->info |= GI_DONE | GI_MINUSMINUS;
-			return EOF;
-		}
-		if (argv[go->optind - 1][0] == '+')
-			go->info |= GI_PLUS;
+		go->info |= flag == '-' ? GI_MINUS : GI_PLUS;
 	}
 	go->p++;
 	if (c == '?' || c == ':' || c == ';' || c == ',' || c == '#'
 	    || !(o = strchr(options, c)))
 	{
-		go->info &= ~GI_PLUS;
 		go->info |= GI_DONE;
 		if (options[0] == ':') {
 			go->buf[0] = c;
 			go->optarg = go->buf;
 		} else {
-			shellf("%s: -%c: bad option\n", argv[0], c);
-			shf_flush(shl_out);
+			warningf(TRUE, "%s: -%c: bad option\n",
+				argv[0], c);
 			if (go->flags & GF_ERROR)
-				errorf((char *) 0);
+				bi_errorf((char *) 0);
 		}
 		return '?';
 	}
@@ -696,17 +772,16 @@ ksh_getopt(argv, go, options)
 		else if (*o == ';')
 			go->optarg = (char *) 0;
 		else {
-			go->info &= ~GI_PLUS;
 			go->info |= GI_DONE;
 			if (options[0] == ':') {
 				go->buf[0] = c;
 				go->optarg = go->buf;
 				return ':';
 			}
-			shellf("%s: -`%c' requires argument\n", argv[0], c);
-			shf_flush(shl_out);
+			warningf(TRUE, "%s: -`%c' requires argument\n",
+				argv[0], c);
 			if (go->flags & GF_ERROR)
-				errorf((char *) 0);
+				bi_errorf((char *) 0);
 			return '?';
 		}
 		go->p = 0;
@@ -744,20 +819,133 @@ void
 print_value_quoted(s)
 	char *s;
 {
+	char *p;
 	int inquote = 0;
 
-	for (; *s; s++) {
-	    if (*s == '\'') {
-		shprintf("'\\'" + 1 - inquote);
-		inquote = 0;
-	    } else {
-		if (!inquote) {
-		    shprintf("'", shl_stdout);
-		    inquote = 1;
+	/* Test if any quotes are needed */
+	for (p = s; *p; p++)
+		if (!letnum(*p) && *p != '/')
+			break;
+	if (!*p) {
+		shprintf("%s", s);
+		return;
+	}
+	for (p = s; *p; p++) {
+		if (*p == '\'') {
+			shprintf("'\\'" + 1 - inquote);
+			inquote = 0;
+		} else {
+			if (!inquote) {
+				shprintf("'");
+				inquote = 1;
+			}
+			shf_putc(*p, shl_stdout);
 		}
-		shf_putc(*s, shl_stdout);
-	    }
 	}
 	if (inquote)
-	    shprintf("'", shl_stdout);
+		shprintf("'");
+}
+
+/* Print things in columns and rows - func() is called to format the ith
+ * element
+ */
+void
+print_columns(shf, n, func, arg, max_width)
+	struct shf *shf;
+	int n;
+	char *(*func) ARGS((void *, int, char *, int));
+	void *arg;
+	int max_width;
+{
+	char *str = (char *) alloc(max_width + 1, ATEMP);
+	int i;
+	int r, c;
+	int rows, cols;
+	int nspace;
+
+	/* max_width + 1 for the space.  Note that no space
+	 * is printed after the last column to avoid problems
+	 * with terminals that have auto-wrap.
+	 */
+	cols = x_cols / (max_width + 1);
+	if (!cols)
+		cols = 1;
+	rows = (n + cols - 1) / cols;
+	if (n && cols > rows) {
+		int tmp = rows;
+
+		rows = cols;
+		cols = tmp;
+		if (rows > n)
+			rows = n;
+	}
+
+	nspace = (x_cols - max_width * cols) / cols;
+	if (nspace <= 0)
+		nspace = 1;
+	for (r = 0; r < rows; r++) {
+		for (c = 0; c < cols; c++) {
+			i = c * rows + r;
+			if (i < n) {
+				shf_fprintf(shf, "%-*s",
+					max_width,
+					(*func)(arg, i, str, max_width + 1));
+				if (c + 1 < cols)
+					shf_fprintf(shf, "%*s", nspace, null);
+			}
+		}
+		shf_putchar('\n', shf);
+	}
+	afree(str, ATEMP);
+}
+
+/* Strip any nul bytes from buf - returns new length (nbytes - # of nuls) */
+int
+strip_nuls(buf, nbytes)
+	char *buf;
+	int nbytes;
+{
+	char *dst;
+
+	/* nbytes check because some systems (older freebsd's) have a buggy
+	 * memchr()
+	 */
+	if (nbytes && (dst = memchr(buf, '\0', nbytes))) {
+		char *end = buf + nbytes;
+		char *p, *q;
+
+		for (p = dst; p < end; p = q) {
+			/* skip a block of nulls */
+			while (++p < end && *p == '\0')
+				;
+			/* find end of non-null block */
+			if (!(q = memchr(p, '\0', end - p)))
+				q = end;
+			memmove(dst, p, q - p);
+			dst += q - p;
+		}
+		*dst = '\0';
+		return dst - buf;
+	}
+	return nbytes;
+}
+
+/* Copy at most dsize-1 bytes from src to dst, ensuring dst is null terminated.
+ * Returns dst.
+ */
+char *
+str_zcpy(dst, src, dsize)
+	char *dst;
+	char *src;
+	int dsize;
+{
+	if (dsize > 0) {
+		int len = strlen(src);
+
+		if (len >= dsize)
+			len = dsize - 1;
+		memcpy(dst, src, len);
+		dst[len] = '\0';
+	}
+	return dst;
 }
