@@ -2,12 +2,7 @@
  * shell parser (C version)
  */
 
-#if !defined(lint) && !defined(no_RCSids)
-static char *RCSid = "$Id: syn.c,v 1.3 1994/05/31 13:34:34 michael Exp $";
-#endif
-
 #include "sh.h"
-#include "expand.h"
 
 struct multiline_state {
 	int	on;		/* set in multiline commands (\n becomes ;) */
@@ -33,14 +28,16 @@ static char **	wordlist	ARGS((void));
 static struct op *block		ARGS((int type, struct op *t1, struct op *t2,
 				      char **wp));
 static struct op *newtp		ARGS((int type));
-static void	syntaxerr	ARGS((void)) GCC_FA_NORETURN;
+static void	syntaxerr	ARGS((void)) GCC_FUNC_ATTR(noreturn);
 static void	multiline_push ARGS((struct multiline_state *save, int tok));
 static void	multiline_pop ARGS((struct multiline_state *saved));
 static int	assign_command ARGS((char *s));
+#ifdef KSH
 static void	db_parse	ARGS((XPtrV *argsp, XPtrV *varps));
 static void	db_oaexpr	ARGS((XPtrV *argsp, XPtrV *varps));
 static void	db_nexpr	ARGS((XPtrV *argsp, XPtrV *varps));
 static void	db_primary	ARGS((XPtrV *argsp, XPtrV *varps));
+#endif /* KSH */
 
 static	struct	op	*outtree; /* yyparse output */
 
@@ -64,8 +61,10 @@ yyparse()
 	if ((tpeek(KEYWORD|ALIAS|VARASN)) == 0) /* EOF */
 		outtree = newtp(TEOF);
 	else {
+		int c;
 		outtree = c_list();
-		musthave('\n', 0);
+		if ((c = token(0)) != '\n' && c != 0)
+			syntaxerr();
 	}
 }
 
@@ -116,15 +115,17 @@ c_list()
 
 	t = andor();
 	if (t != NULL) {
-		while ((c = token(0)) == ';' || c == '&' ||
+		while ((c = token(0)) == ';' || c == '&' || c == COPROC ||
 		       (c == '\n' && (multiline.on || source->type == SSTRING
 				      || source->type == SALIAS)))
 		{
-			if (c == '&') {
+			if (c == '&' || c == COPROC) {
+				int type = c == '&' ? TASYNC : TCOPROC;
 				if (tl)
-					tl->right = block(TASYNC, tl->right, NOBLOCK, NOWORDS);
+					tl->right = block(type, tl->right,
+							  NOBLOCK, NOWORDS);
 				else
-					t = block(TASYNC, t, NOBLOCK, NOWORDS);
+					t = block(type, t, NOBLOCK, NOWORDS);
 			}
 			if ((p = andor()) == NULL)
 				return (t);
@@ -143,20 +144,23 @@ synio(cf)
 	int cf;
 {
 	register struct ioword *iop;
+	int ishere;
 
 	if (tpeek(cf) != REDIR)
 		return NULL;
 	ACCEPT;
 	iop = yylval.iop;
-	musthave(LWORD, 0);
-	iop->name = yylval.cp;
-	if ((iop->flag&IOTYPE) == IOHERE) {
+	ishere = (iop->flag&IOTYPE) == IOHERE;
+	musthave(LWORD, ishere ? HEREDELIM : 0);
+	if (ishere) {
+		iop->delim = yylval.cp;
 		if (*ident != 0) /* unquoted */
 			iop->flag |= IOEVAL;
 		if (herep >= &heres[HERES])
 			yyerror("too many <<'s\n");
 		*herep++ = iop;
-	}
+	} else
+		iop->name = yylval.cp;
 	return iop;
 }
 
@@ -201,9 +205,6 @@ get_command(cf)
 		cf = CONTIN;
 	syniocf = KEYWORD|ALIAS;
 	switch (c = token(cf|KEYWORD|ALIAS|VARASN)) {
-	  case 0:
-		syntaxerr();
-
 	  default:
 		REJECT;
 		return NULL; /* empty line */
@@ -215,7 +216,7 @@ get_command(cf)
 		t = newtp(TCOM);
 		while (1) {
 			cf = (t->evalflags ? ARRAYVAR : 0)
-			     | (XPsize(args) == 0 ? ALIAS|VARASN : 0);
+			     | (XPsize(args) == 0 ? ALIAS|VARASN : CMDWORD);
 			switch (tpeek(cf)) {
 			  case REDIR:
 				if (iopn >= NUFILE)
@@ -287,12 +288,14 @@ get_command(cf)
 		break;
 	  }
 
+#ifdef KSH
 	  case DBRACKET: /* [[ .. ]] */
 		syniocf &= ~(KEYWORD|ALIAS);
 		t = newtp(TDBRACKET);
 		ACCEPT;
 		db_parse(&args, &vars);
 		break;
+#endif /* KSH */
 
 	  case FOR:
 	  case SELECT:
@@ -393,10 +396,19 @@ dogroup()
 	register struct op *list;
 
 	c = token(CONTIN|KEYWORD|ALIAS);
-	if (c != DO)
+	/* A {...} can be used instead of do...done for for/select loops
+	 * but not for while/until loops - we don't need to check if it
+	 * is a while loop because it would have been parsed as part of
+	 * the conditional command list...
+	 */
+	if (c == DO)
+		c = DONE;
+	else if (c == '{')
+		c = '}';
+	else
 		syntaxerr();
 	list = c_list();
-	musthave(DONE, KEYWORD|ALIAS);
+	musthave(c, KEYWORD|ALIAS);
 	return list;
 }
 
@@ -490,7 +502,7 @@ function_body(name, ksh_func)
 	struct op *t;
 	int old_func_parse;
 
-	Xinit(xs, xp, 16);
+	Xinit(xs, xp, 16, ATEMP);
 	for (p = name; ; ) {
 		if ((*p == EOS && Xlength(xs, xp) == 0)
 		    || (*p != EOS && *p != CHAR && *p != QCHAR
@@ -546,7 +558,8 @@ wordlist()
 
 	XPinit(args, 16);
 	if ((c = token(CONTIN|KEYWORD|ALIAS)) != IN) {
-		REJECT;
+		if (c != ';') /* non-POSIX, but at&t ksh accepts a ; here */
+			REJECT;
 		return NULL;
 	}
 	while ((c = token(0)) == LWORD)
@@ -595,7 +608,9 @@ const	struct tokeninfo {
 	{ "case",	CASE,	TRUE },
 	{ "esac",	ESAC,	TRUE },
 	{ "for",	FOR,	TRUE },
+#ifdef KSH
 	{ "select",	SELECT,	TRUE },
+#endif /* KSH */
 	{ "while",	WHILE,	TRUE },
 	{ "until",	UNTIL,	TRUE },
 	{ "do",		DO,	TRUE },
@@ -606,12 +621,15 @@ const	struct tokeninfo {
 	{ "{",		'{',	TRUE },
 	{ "}",		'}',	TRUE },
 	{ "!",		BANG,	TRUE },
+#ifdef KSH
 	{ "[[",		DBRACKET, TRUE },
+#endif /* KSH */
 	/* Lexical tokens (0[EOF], LWORD and REDIR handled specially) */
 	{ "&&",		LOGAND,	FALSE },
 	{ "||",		LOGOR,	FALSE },
 	{ ";;",		BREAK,	FALSE },
 	{ "((",		MDPAREN, FALSE },
+	{ "|&",		COPROC,	FALSE },
 	/* and some special cases... */
 	{ "newline",	'\n',	FALSE },
 	{ 0 }
@@ -623,6 +641,7 @@ initkeywords()
 	register struct tokeninfo const *tt;
 	register struct tbl *p;
 
+	tinit(&keywords, APERM, 32); /* must be 2^n (currently 20 keywords) */
 	for (tt = tokentab; tt->name; tt++) {
 		if (tt->reserved) {
 			p = tenter(&keywords, tt->name, hash(tt->name));
@@ -662,7 +681,6 @@ syntaxerr()
 		break;
 
 	case REDIR:
-		yylval.iop->name = (char *) 0;
 		snptreef(s = redir, sizeof(redir), "%R", yylval.iop);
 		break;
 
@@ -728,8 +746,6 @@ compile(s)
 	herep = heres;
 	source = s;
 	yyparse();
-	if (s->type == STTY || s->type == SFILE || s->type == SHIST)
-		s->str = null;	/* line is not preserved */
 	return outtree;
 }
 
@@ -758,7 +774,8 @@ assign_command(s)
 }
 
 
-#define db_makevar(c)	(db_fakearg[1] = (c), wdcopy(db_fakearg, ATEMP))
+#ifdef KSH
+# define db_makevar(c)	(db_fakearg[1] = (c), wdcopy(db_fakearg, ATEMP))
 
 /* used by db_makevar() */
 static char db_fakearg[3] = { CHAR, 'x', EOS };
@@ -886,3 +903,4 @@ db_primary(argsp, varsp)
 	} else
 		syntaxerr();
 }
+#endif /* KSH */
