@@ -7,14 +7,9 @@
 #include "config.h"
 #ifdef VI
 
-#if !defined(lint) && !defined(no_RCSids)
-static char *RCSid = "$Id: vi.c,v 1.3 1994/05/31 13:34:34 michael Exp michael $";
-#endif
-
 #include "sh.h"
 #include <ctype.h>
 #include "ksh_stat.h"		/* completion */
-#include "expand.h"
 #include "edit.h"
 
 #define CMDLEN		1024
@@ -57,6 +52,7 @@ static int 	glob_word ARGS((Glob *g, int command));
 static int	expand_word ARGS((int command));
 static int	complete_word ARGS((int command));
 static int	print_expansions ARGS((int command));
+static int 	char_len ARGS((int c));
 static void 	x_vi_zotc ARGS((int c));
 
 #define C_	0x1
@@ -65,7 +61,6 @@ static void 	x_vi_zotc ARGS((int c));
 #define X_	0x8
 #define U_	0x10
 #define B_	0x20
-#define O_	0x40
 #define S_	0x80
 
 #define isbad(c)	(classify[(c)&0x7f]&B_)
@@ -73,28 +68,27 @@ static void 	x_vi_zotc ARGS((int c));
 #define ismove(c)	(classify[(c)&0x7f]&M_)
 #define isextend(c)	(classify[(c)&0x7f]&E_)
 #define islong(c)	(classify[(c)&0x7f]&X_)
-#define ismeta(c)	(classify[(c)&0x7f]&O_)
 #define isundoable(c)	(!(classify[(c)&0x7f]&U_))
 #define issrch(c)	(classify[(c)&0x7f]&S_)
 
 char	classify[128] = {
-	/*	0	1	2	3	4	5	6	7  */
-/*  0- */	B_,	0,	0,	0,	0,	C_|U_,	O_,	0,
-/*  1- */	C_|M_,	0,	O_,	0,	C_|U_,	O_,	O_|C_,	0,
-/*  2- */	O_|C_,	0,	C_|U_,	0,	0,	0,	0,	0,
-/*  3- */	0,	0,	O_,	C_,	0,	0,	0,	0,
-/*  4- */	M_,	0,	0,	C_,	M_,	M_,	0,	0,
-/*  5- */	0,	0,	C_,	C_,	M_,	C_,	0,	C_|S_,
-/*  6- */	M_,	0,	0,	0,	0,	0,	0,	0,
-/*  7- */	0,	0,	0,	M_,	0,	C_,	0,	C_|S_,
-/* 10- */	0,	C_,	M_,	C_,	C_,	M_,	M_|X_,	C_,
-/* 11- */	0,	C_,	0,	0,	0,	0,	C_,	0,
-/* 12- */	C_,	0,	C_,	C_,	M_|X_,	0,	0,	M_,
-/* 13- */	C_,	C_,	0,	0,	C_,	0,	M_,	C_,
-/* 14- */	0,	C_,	M_,	E_,	E_,	M_,	M_|X_,	0,
-/* 15- */	M_,	C_,	C_,	C_,	M_,	0,	C_,	0,
-/* 16- */	C_,	0,	X_,	C_,	M_|X_,	C_|U_,	0,	M_,
-/* 17- */	C_,	E_,	0,	0,	M_,	0,	C_,	0
+	     /*	0	1	2	3	4	5	6	7  */
+   /*   0 */	B_,	0,	0,	0,	0,	C_|U_,	C_|U_,	0,
+   /*  01 */	C_|M_,	0,	0,	0,	C_|U_,	0,	C_,	0,
+   /*  02 */	C_,	0,	C_|U_,	0,	0,	0,	0,	0,
+   /*  03 */	0,	0,	C_,	0,	0,	0,	0,	0,
+   /*  04 */	M_,	0,	0,	C_,	M_,	M_,	0,	0,
+   /*  05 */	0,	0,	C_,	C_,	M_,	C_,	0,	C_|S_,
+   /*  06 */	M_,	0,	0,	0,	0,	0,	0,	0,
+   /*  07 */	0,	0,	0,	M_,	0,	C_,	0,	C_|S_,
+   /* 010 */	0,	C_,	M_,	C_,	C_,	M_,	M_|X_,	C_,
+   /* 011 */	0,	C_,	0,	0,	0,	0,	C_,	0,
+   /* 012 */	C_,	0,	C_,	C_,	M_|X_,	0,	0,	M_,
+   /* 013 */	C_,	C_,	0,	0,	C_,	0,	M_,	C_,
+   /* 014 */	0,	C_,	M_,	E_,	E_,	M_,	M_|X_,	C_,
+   /* 015 */	M_,	C_,	C_,	C_,	M_,	0,	C_,	0,
+   /* 016 */	C_,	0,	X_,	C_,	M_|X_,	C_|U_,	C_,	M_,
+   /* 017 */	C_,	E_,	0,	0,	M_,	0,	C_,	0
 };
 
 #define MAXVICMD	3
@@ -284,17 +278,14 @@ vi_hook(ch)
 
 	case VSEARCH:
 		if (ch == '\r' || ch == '\n' /*|| ch == Ctrl('[')*/ ) {
+			restore_cbuf();
 			locpat[srchlen] = '\0';
 			(void) strcpy(srchpat, locpat);
 			state = VCMD;
 		} else if (ch == edchars.erase || ch == Ctrl('h')) {
 			if (srchlen != 0) {
 				srchlen--;
-				if (locpat[srchlen] < ' ' ||
-						locpat[srchlen] == 0x7f) {
-					es->linelen--;
-				}
-				es->linelen--;
+				es->linelen -= char_len((unsigned char) locpat[srchlen]);
 				es->cursor = es->linelen;
 				refresh(0);
 				return 0;
@@ -316,11 +307,8 @@ vi_hook(ch)
 				n--;
 			while (n > 0 && !isspace(locpat[n - 1]))
 				n--;
-			for (i = srchlen; --i >= n; ) {
-				if (locpat[i] < ' ' || locpat[i] == 0x7f)
-					es->linelen--;
-				es->linelen--;
-			}
+			for (i = srchlen; --i >= n; )
+				es->linelen -= char_len((unsigned char) locpat[i]);
 			srchlen = n;
 			es->cursor = es->linelen;
 			refresh(0);
@@ -330,6 +318,11 @@ vi_hook(ch)
 				x_putc(BEL);
 			else {
 				locpat[srchlen++] = ch;
+				if ((ch & 0x80) && Flag(FVISHOW8)) {
+					es->cbuf[es->linelen++] = 'M';
+					es->cbuf[es->linelen++] = '-';
+					ch &= 0x7f;
+				}
 				if (ch < ' ' || ch == 0x7f) {
 					es->cbuf[es->linelen++] = '^';
 					es->cbuf[es->linelen++] = ch ^ '@';
@@ -358,6 +351,9 @@ vi_hook(ch)
 		case 1:
 			refresh(0);
 			return 1;
+		case 2:
+			/* back from a 'v' command - don't redraw the screen */
+			return 1;
 		}
 		break;
 
@@ -386,6 +382,9 @@ vi_hook(ch)
 		case 1:
 			refresh(0);
 			return 1;
+		case 2:
+			/* back from a 'v' command - don't redraw the screen */
+			return 1;
 		}
 		break;
 
@@ -556,9 +555,8 @@ vi_cmd(argcnt, cmd)
 	struct edstate	*t;
 
 	if (argcnt == 0) {
-		if (*cmd == 'G')
-			argcnt = hlast + 1;
-		else if (*cmd != '_' && *cmd != '|')
+		if (*cmd != 'G' && *cmd != 'g' && *cmd != '_' && *cmd != '|'
+		    && *cmd != 'v')
 			argcnt = 1;
 	}
 
@@ -600,8 +598,8 @@ vi_cmd(argcnt, cmd)
 			break;
 
 		case 'S':
-			del_range(0, es->linelen);
-			es->cursor = 0;
+			es->cursor = domove(1, "^", 1);
+			del_range(es->cursor, es->linelen);
 			modified = 1; hnum = hlast;
 			insert = INSERT;
 			break;
@@ -613,7 +611,7 @@ vi_cmd(argcnt, cmd)
 		case 'd':
 		case 'y':
 			if (*cmd == cmd[1]) {
-				c1 = 0;
+				c1 = *cmd == 'c' ? domove(1, "^", 1) : 0;
 				c2 = es->linelen;
 			} else if (!ismove(cmd[1]))
 				return -1;
@@ -682,7 +680,15 @@ vi_cmd(argcnt, cmd)
 				es->cursor--;
 			break;
 
+		case 'g':
+			if (!argcnt)
+				argcnt = hlast + 1;
+			/* fall through */
 		case 'G':
+			if (!argcnt)
+				argcnt = 1;
+			else
+				argcnt = hlast - (source->line - argcnt);
 			if (grabhist(modified, argcnt - 1) < 0)
 				return -1;
 			else {
@@ -698,7 +704,7 @@ vi_cmd(argcnt, cmd)
 
 		case 'I':
 			modified = 1; hnum = hlast;
-			es->cursor = 0;
+			es->cursor = domove(1, "^", 1);
 			insert = INSERT;
 			break;
 
@@ -749,6 +755,25 @@ vi_cmd(argcnt, cmd)
 			insert = INSERT;
 			break;
 
+		case 'v':
+			if (es->linelen == 0)
+				return -1;
+			if (!argcnt) {
+				if (modified) {
+					es->cbuf[es->linelen] = '\0';
+					source->line++;
+					histsave(source->line, es->cbuf, 1);
+				} else
+					argcnt = source->line + 1
+						- (hlast - hnum);
+			}
+			shf_snprintf(es->cbuf, es->cbufsize,
+					argcnt ? "%s %d" : "%s",
+					"fc -e ${VISUAL:-${EDITOR:-vi}} --",
+					argcnt);
+			es->linelen = strlen(es->cbuf);
+			return 2;
+
 		case 'x':
 			if (es->linelen == 0)
 				return -1;
@@ -809,7 +834,7 @@ vi_cmd(argcnt, cmd)
 			}
 			break;
 		case '_': {
-			int	space;
+			int	inspace;
 			char	*p, *sp;
 
 			if (histnum(-1) < 0)
@@ -830,12 +855,12 @@ vi_cmd(argcnt, cmd)
 				sp = p;
 			} else {
 				sp = p;
-				space = 0;
+				inspace = 0;
 				while (*p) {
 					if (issp(*p))
-						space = 1;
-					else if (space) {
-						space = 0;
+						inspace = 1;
+					else if (inspace) {
+						inspace = 0;
 						sp = p;
 					}
 					p++;
@@ -849,7 +874,7 @@ vi_cmd(argcnt, cmd)
 				argcnt++;
 				p++;
 			}
-			if (putbuf(" ", 1, 0) != 0)
+			if (putbuf(space, 1, 0) != 0)
 				argcnt = -1;
 			else if (putbuf(sp, argcnt, 0) != 0)
 				argcnt = -1;
@@ -864,21 +889,24 @@ vi_cmd(argcnt, cmd)
 
 		case '~': {
 			char	*p;
+			int	i;
 
 			if (es->linelen == 0)
 				return -1;
-			p = &es->cbuf[es->cursor];
-			if (islower(*p)) {
-				modified = 1; hnum = hlast;
-				*p = toupper(*p);
-			} else if (isupper(*p)) {
-				modified = 1; hnum = hlast;
-				*p = tolower(*p);
-			}
-			if (es->cursor < es->linelen - 1)
-				es->cursor++;
+			for (i = 0; i < argcnt; i++) {
+				p = &es->cbuf[es->cursor];
+				if (islower(*p)) {
+					modified = 1; hnum = hlast;
+					*p = toupper(*p);
+				} else if (isupper(*p)) {
+					modified = 1; hnum = hlast;
+					*p = tolower(*p);
+				}
+				if (es->cursor < es->linelen - 1)
+					es->cursor++;
 			}
 			break;
+			}
 
 		case '#':
 			es->cursor = 0;
@@ -891,10 +919,10 @@ vi_cmd(argcnt, cmd)
 			print_expansions(1);
 			break;
 
-		case Ctrl('['):			/* at&t ksh */
+		/* case Ctrl('['):	some annoying at&t ksh's */
 		case '\\':			/* at&t ksh */
 		case Ctrl('f'):			/* Nonstandard vi/ksh */
-			complete_word(0);
+			complete_word(1);
 			break;
 
 
@@ -1419,12 +1447,13 @@ grabhist(save, n)
 	}
 	(void) histnum(n);
 	if ((hptr = *histpos()) == NULL) {
-		shellf("grabhist: bad history array\n");
+		internal_errorf(0, "grabhist: bad history array");
 		return -1;
 	}
 	if (save)
 		save_cbuf();
-	es->linelen = strlen(hptr);
+	if ((es->linelen = strlen(hptr)) >= es->cbufsize - 1)
+		es->linelen = es->cbufsize - 1;
 	memmove(es->cbuf, hptr, es->linelen);
 	es->cursor = 0;
 	return 0;
@@ -1437,6 +1466,7 @@ grabsearch(save, start, fwd, pat)
 {
 	char	*hptr;
 	int	hist;
+	int	anchored;
 
 	if ((start == 0 && fwd == 0) || (start >= hlast-1 && fwd == 1))
 		return -1;
@@ -1444,9 +1474,10 @@ grabsearch(save, start, fwd, pat)
 		start++;
 	else
 		start--;
-	if ((hist = findhist(start, fwd, pat)) < 0) {
+	anchored = *pat == '^' ? (++pat, 1) : 0;
+	if ((hist = findhist(start, fwd, pat, anchored)) < 0) {
 		/* if (start != 0 && fwd && match(holdbuf, pat) >= 0) { */
-		if (start != 0 && fwd && strcmp(holdbuf, pat) >= 0) {
+		if (start != 0 && fwd && FILECMP(holdbuf, pat) >= 0) {
 			restore_cbuf();
 			return 0;
 		} else
@@ -1456,7 +1487,8 @@ grabsearch(save, start, fwd, pat)
 		save_cbuf();
 	histnum(hist);
 	hptr = *histpos();
-	es->linelen = strlen(hptr);
+	if ((es->linelen = strlen(hptr)) >= es->cbufsize - 1)
+		es->linelen = es->cbufsize - 1;
 	memmove(es->cbuf, hptr, es->linelen);
 	es->cursor = 0;
 	return hist;
@@ -1497,7 +1529,7 @@ outofwin()
 	col = 0;
 	cur = es->winleft;
 	while (cur < es->cursor)
-		col = newcol(es->cbuf[cur++], col);
+		col = newcol((unsigned char) es->cbuf[cur++], col);
 	if (col >= winwidth)
 		return 1;
 	return 0;
@@ -1519,10 +1551,11 @@ rewindow()
 			holdcur2 = tcur;
 			holdcol2 = tcol;
 		}
-		tcol = newcol(es->cbuf[tcur++], tcol);
+		tcol = newcol((unsigned char) es->cbuf[tcur++], tcol);
 	}
 	while (tcol - holdcol1 > winwidth / 2)
-		holdcol1 = newcol(es->cbuf[holdcur1++], holdcol1);
+		holdcol1 = newcol((unsigned char) es->cbuf[holdcur1++],
+				  holdcol1);
 	es->winleft = holdcur1;
 }
 
@@ -1530,13 +1563,9 @@ static int
 newcol(ch, col)
 	int	ch, col;
 {
-	if (ch < ' ' || ch == 0x7f) {
-		if (ch == '\t')
-			return (col | 7) + 1;
-		else
-			return col + 2;
-	} else
-		return col + 1;
+	if (ch == '\t')
+		return (col | 7) + 1;
+	return col + char_len(ch);
 }
 
 static void
@@ -1544,6 +1573,7 @@ display(wb1, wb2, leftside)
 	char	*wb1, *wb2;
 	int	leftside;
 {
+	unsigned char ch;
 	char	*twb1, *twb2, mc;
 	int	cur, col, cnt;
 	int	UNINITIALIZED(ncol);
@@ -1556,21 +1586,31 @@ display(wb1, wb2, leftside)
 	while (col < winwidth && cur < es->linelen) {
 		if (cur == es->cursor && leftside)
 			ncol = col + pwidth;
-		if (es->cbuf[cur] < ' ' || es->cbuf[cur] == 0x7f) {
-			if (es->cbuf[cur] == '\t') {
-				do {
-					*twb1++ = ' ';
-				} while (++col < winwidth && (col & 7) != 0);
-			} else {
-				*twb1++ = '^';
+		if ((ch = es->cbuf[cur]) == '\t') {
+			do {
+				*twb1++ = ' ';
+			} while (++col < winwidth && (col & 7) != 0);
+		} else {
+			if ((ch & 0x80) && Flag(FVISHOW8)) {
+				*twb1++ = 'M';
 				if (++col < winwidth) {
-					*twb1++ = es->cbuf[cur] ^ '@';
+					*twb1++ = '-';
+					col++;
+				}
+				ch &= 0x7f;
+			}
+			if (col < winwidth) {
+				if (ch < ' ' || ch == 0x7f) {
+					*twb1++ = '^';
+					if (++col < winwidth) {
+						*twb1++ = es->cbuf[cur] ^ '@';
+						col++;
+					}
+				} else {
+					*twb1++ = es->cbuf[cur];
 					col++;
 				}
 			}
-		} else {
-			*twb1++ = es->cbuf[cur];
-			col++;
 		}
 		if (cur == es->cursor && !leftside)
 			ncol = col + pwidth - 1;
@@ -1661,13 +1701,12 @@ x_vi(buf, len)
 				/* pretend we got an interrupt */
 				x_vi_zotc(c);
 				x_flush();
-				shf_flush(shl_out);
 				trapsig(c == edchars.intr ? SIGINT : SIGQUIT);
-				runtraps(TRUE);
-				continue;
-			} else if (c == Ctrl('d')) {
+				x_mode(FALSE);
+				unwind(LSHELL);
+			} else if (c == edchars.eof) {
 				if (es->linelen == 0) {
-					x_putc('^'); x_putc('D');
+					x_vi_zotc(edchars.eof);
 					c = -1;
 					break;
 				}
@@ -1704,12 +1743,12 @@ globstr(cp)
 	struct source	*s, *sold;
 
 	sold = source;
-	s = pushs(SWSTR);
+	s = pushs(SWSTR, ATEMP);
 	s->str = (char *) cp;
 	source = s;
 	if (yylex(ONEWORD) != LWORD) {
 		source = sold;
-		shellf("eval:substitute error\n");
+		internal_errorf(0, "globstr: substitute error");
 		return (char **) 0;
 	}
 	source = sold;
@@ -1805,9 +1844,10 @@ glob_word(g, command)
 
 	/* restore undo buffer that we used temporarily */
 	memmove(toglob, es->cbuf, es->linelen);
-	return 0;
+	return rval;
 }
 
+/* replace word with all expansions (ie, expand word*) */
 static int
 expand_word(command)
 	int command;
@@ -1840,13 +1880,13 @@ expand_word(command)
 		}
 		if (*++g.result == (char *) 0)
 			break;
-		if (putbuf(" ", 1, 0) != 0) {
+		if (putbuf(space, 1, 0) != 0) {
 			rval = -1;
 			break;
 		}
 	}
 	if (g.addspace && rval == 0)
-		rval = putbuf(" ", 1, 0);
+		rval = putbuf(space, 1, 0);
 	modified = 1; hnum = hlast;
 	insert = INSERT;
 	refresh(0);
@@ -1902,9 +1942,7 @@ complete_word(command)
 	expanded = COMPLETE;
 	del_range(g.start, g.end);
 	es->cursor = g.start;
-	if (g.result == (char **)0)
-		rval = -1;
-	else {
+	{
 		/* find the shortest match */
 		char	*save = g.result[0], *p;
 		len = strlen(g.result[0]);
@@ -1913,8 +1951,8 @@ complete_word(command)
 		for (i=1; (p = g.result[i]); i++) {
 			int j;
 
-			for (j=0; j<len; j++)
-				if (p[j] != save[j]) {
+			for (j = 0; j < len; j++)
+				if (FILECHCONV(p[j]) != FILECHCONV(save[j])) {
 					len = j;
 					break;
 				}
@@ -1922,16 +1960,12 @@ complete_word(command)
 		if (putbuf(g.result[0], len, 0) != 0) 
 			rval = -1;
 	}
-	/* Check to see if it's a directory. */
-	if (g.result[1] == 0) {
-		if (stat(g.result[0], &s) < 0)
-			rval = -1;
-		else {
-			if (S_ISDIR(s.st_mode))
-				putbuf("/", 1, 0);
-			else
-				putbuf(" ", 1, 0);
-		}
+	/* Check to see if it's a directory (len == 0 probably isn't needed) */
+	if (g.result[1] == 0 && (len == 0 || !ISDIRSEP(g.result[0][len - 1]))) {
+		if (stat(g.result[0], &s) >= 0 && S_ISDIR(s.st_mode))
+			putbuf(slash, 1, 0);
+		else
+			putbuf(space, 1, 0);
 	}
 	modified = 1; hnum = hlast;
 	insert = INSERT;
@@ -1957,7 +1991,7 @@ print_expansions(command)
 		x_putc(' ');
 	while (*p && **p) {
 		char 	*q;
-		if ((q = strrchr(*p, '/')) == 0)
+		if ((q = strrchr_dirsep(*p)) == 0)
 			x_puts(*p);
 		else
 			x_puts(++q);
@@ -1970,12 +2004,31 @@ print_expansions(command)
 	return 0;
 }
 
-/* Same as x_zotc(emacs.c), but no tab wierdness */
+/* How long is char when displayed (not counting tabs) */
+static int
+char_len(c)
+	int c;
+{
+	int len = 1;
+
+	if ((c & 0x80) && Flag(FVISHOW8)) {
+		len += 2;
+		c &= 0x7f;
+	}
+	if (c < ' ' || c == 0x7f)
+		len++;
+	return len;
+}
+
+/* Similar to x_zotc(emacs.c), but no tab wierdness */
 static void
 x_vi_zotc(c)
 	int c;
 {
-	c &= 0x7f;
+	if (Flag(FVISHOW8) && (c & 0x80)) {
+		x_puts("M-");
+		c &= 0x7f;
+	}
 	if (c < ' ' || c == 0x7f) {
 		x_putc('^');
 		c ^= '@';
